/************************************************************************
	Copyright (C) 2010 Robert Scott.
	
	This file if part of envi_feed.
	
    envi_feed is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    envi_feed is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
	
*************************************************************************/

#include "config.h"
#include <zmq.hpp>
#include <boost/asio.hpp>
#include <boost/asio/serial_port.hpp>

#include <log4cplus/logger.h>
#include <log4cplus/loglevel.h>
#include <log4cplus/configurator.h>
#include <log4cplus/layout.h>

#ifdef _WIN32
#include "XGetopt.h"
#endif

#include "envi.h"

const std::string version(PACKAGE_VERSION);


bool load_config(std::string filename, std::string & port, std::string &logger_code, std::vector<std::string> & endpoint_list);
void process_messages(boost::asio::serial_port * port,  const std::string & logger_code, std::vector<std::string> & endpoints );
#ifdef _WIN32
void service_main(int argc, char **argv);
bool install_service();
bool remove_service();
extern std::string service_name;
#endif 

#ifndef _WIN32
sig_atomic_t run = 1;
#else
int run = 1;
#endif

boost::asio::serial_port * port;
envi envi_device;
std::string logger_code;
std::vector<std::string> endpoints;

void cleanup()
{
	port->close();
	log4cplus::Logger logger=log4cplus::Logger::getRoot();
	LOG4CPLUS_INFO(logger, "closed");
	logger.shutdown();
}

#ifndef _WIN32
void sig_handler(int signum)
{
	log4cplus::Logger logger=log4cplus::Logger::getRoot();
	bool terminate = false;
	switch (signum) {
		case SIGHUP : 
			LOG4CPLUS_INFO(logger, "received SIGHUP");
			break;
		case SIGTERM:
			LOG4CPLUS_INFO(logger, "received SIGTERM");
			terminate=true;
			break;
		case SIGINT:
			LOG4CPLUS_INFO(logger, "received SIGINT");
			terminate=true;
			break;
		default:
			std::ostringstream log_message;
			log_message << "unhandled signal (" << strsignal(signum) << ")";
			LOG4CPLUS_INFO(logger, log_message); 
			break;
	}
	if(terminate) {
		LOG4CPLUS_INFO(logger, "closing"); 
		cleanup();
		run=0;
	}
}
#endif





int main(int argc, char **argv)
{
	int opt;
	bool daemonise = false;
	std::string config_file;
	std::string logger_config;
	boost::asio::io_service  io_service;
	boost::asio::streambuf stream_buffer;	

#ifndef _WIN32
	config_file="/etc/envi_feed/config/envi_feed.conf";
	logger_config="/etc/envi_feed/config/logger.conf";
#else
	config_file="C:\\Documents and Settings\\All Users\\Application Data\\envi_feed\\envi_feed.conf";
	logger_config="C:\\Documents and Settings\\All Users\\Application Data\\envi_feed\\logger.conf";
#endif


#ifdef _WIN32
	while((opt=getopt(argc, argv, "dirf:"))!=-1){
#else
	while((opt=getopt(argc, argv, "df:"))!=-1){
#endif
		switch(opt) {
			case 'd':
				daemonise = true;
				break; 
			case 'f':
				config_file = optarg;
				break;
			case 'l':
				logger_config = optarg;
				break;
#ifdef _WIN32
			case 'i':
				if(!install_service()) {
					std::cout << "service install failed" << std::endl;
					exit(EXIT_FAILURE);
				}
				std::cout << "service installed successfully" << std::endl;
				exit(EXIT_SUCCESS);
				break;
			case 'r':
				if(!remove_service()) {
					std::cout << "service remove failed" << std::endl;
					exit(EXIT_FAILURE);
				}
				std::cout << "service removed successfully" << std::endl;
				exit(EXIT_SUCCESS);
				break;
#endif
		}
	}
	log4cplus::PropertyConfigurator::doConfigure(logger_config);
	log4cplus::Logger logger=log4cplus::Logger::getRoot();
	std::ostringstream log_message;
	log_message << "version " << version;
	LOG4CPLUS_INFO(logger, log_message.str());
	LOG4CPLUS_INFO(logger, "starting");
	

	std::string port_name;

	if(!load_config(config_file, port_name, logger_code, endpoints)) {
		cleanup();
		exit(EXIT_FAILURE);
	}
	log_message.str("");
	log_message << "using port " <<  port_name;
	LOG4CPLUS_INFO(logger, log_message.str());	
	
	for(std::vector<std::string>::iterator it = endpoints.begin(); it!=endpoints.end();++it) {
		log_message.str("");
		log_message << "endpoint " << *it;
		LOG4CPLUS_INFO(logger,  log_message.str());
	}

	port= new boost::asio::serial_port(io_service);
	try {
		port->open(port_name);
		port->set_option(boost::asio::serial_port_base::baud_rate(57600));
		port->set_option(boost::asio::serial_port_base::character_size(8));
		port->set_option(boost::asio::serial_port_base::parity(boost::asio::serial_port_base::parity::none));
		port->set_option(boost::asio::serial_port_base::stop_bits(boost::asio::serial_port_base::stop_bits::one));
	}
	catch (boost::system::system_error e) {
		log_message.str("");
		log_message << "error opening port " << port << "(" << e.what() << ")";
		LOG4CPLUS_ERROR(logger, log_message.str());
		cleanup();
		exit(EXIT_FAILURE);
	}

#ifdef _WIN32
	if(daemonise) {
		SERVICE_TABLE_ENTRY service_table[2];
		service_table[0].lpServiceName = (LPSTR)service_name.c_str();
		service_table[0].lpServiceProc = (LPSERVICE_MAIN_FUNCTION)service_main;
		service_table[1].lpServiceName = NULL;
		service_table[1].lpServiceProc = NULL;
		LOG4CPLUS_INFO(logger, "starting service");
		StartServiceCtrlDispatcher(service_table);
		exit(EXIT_SUCCESS);
	}
	else {
		process_messages(port, logger_code, endpoints);
	}
#else
	signal(SIGHUP, sig_handler);
	signal(SIGTERM, sig_handler);
	signal(SIGINT, sig_handler);
	signal(SIGQUIT, sig_handler);
	if(daemonise) {
		LOG4CPLUS_INFO(logger, "run as daemon");
		daemon(0, 0);
	}
	process_messages(port, logger_code, endpoints);
#endif
	cleanup();
	exit(EXIT_SUCCESS);
}

